module MACRO_PASS_M

  implicit NONE
  private
  public :: Path_T, First_Path, Last_Path, Macro_Depth
  public :: Add_Path, Macro_Pass

  type :: Path_T
    character(len=1023) :: Path ! A path to search for macros
    integer :: Len = 0
    type(path_t), pointer :: Next => NULL()
  end type Path_T

  type(path_t), pointer :: First_path => NULL(), Last_Path => NULL()

  ! if Macro_Depth /= 0, read from unit U_Input+Macro_Depth.
  integer, save :: Macro_Depth = 0
  integer, parameter :: Max_Depth = 9 ! Maximum CALL + INCLD depth

contains

  ! ---------------------------------------------------  Add_Path  -----
  subroutine Add_Path ( Path )
  ! Add a path to the list of paths to search for macro files
    character(len=*), intent(in) :: Path
    if ( .not. associated(first_path) ) then
      allocate ( first_path )
      first_path%path = '.'
      first_path%len = 1
      last_path => first_path
    end if
    allocate ( last_path%next )
    last_path => last_path%next
    last_path%path = path
    last_path%len = len(path)
  end subroutine Add_Path

  ! -------------------------------------------------  Macro_Pass  -----
  subroutine MACRO_PASS

  ! Process the input file, expanding macros, and including INCLD and
  ! CALL macros.
  ! Generate branches for CALL entries and DCW's for their args.
  ! Accumulate CALL and INCLD file names.
  ! Generate a LITS record before EX or END or after LTORG.
  ! Insert CALL files before LITS record.

    use ERROR_M, only: ErrCode, BadStatement, DO_ERROR, MacroErr, Overcall
    use INPUT_M, only: INUNIT, READ_LINE
    use IO_UNITS, only: U_INPUT, U_SCR2
    use LEXER, only: LEX, T_NAME
    use MACHINE, only: IO_ERROR
    use OPERAND_M, only: Num_Operands, Operands, K_None, K_Symbolic
    use OP_CODES_M, only: OP_CODES
    use PARSER, only: PARSE

    integer, parameter :: MAX_CALLS = 55

    integer :: ARGS(0:99)      ! ARGS(i) == end of i'th arg of macro or call
    character(5) :: CMD
    character(len=1023) :: FILE ! File name for error messages
    integer :: I
    integer :: IOSTAT
    character(len=6) :: Label   ! from CALL statement
    character(len=400) :: LINE
    character(len=80) :: LineSave
    character(len=3) :: CallQueue(max_calls)   ! CALL held until EX, END, LTORG
    integer :: NumARGS   ! Number of args of CALL or macro
    integer :: NumCALLS
    integer :: SAVE_INUNIT
    logical :: Saved_Line ! Process LineSave
    integer :: Status
    integer :: TheCall   ! Which call is being processed
    character(5) :: WHY  ! Why is record on scratch file?
                         ! '     ' => Directly from input
                         ! 'MACRO' => CALL, INCLD or macro
                         ! 'GEN  ' => Generated line
                         ! 'LITS ' => Time to generate lits

    line = ''
    numCalls = 0
    saved_line = .false.
    save_inunit = inunit ! stdin or a file
    theCall = 0
  o:do
      errCode = ''
      if ( macro_depth == 0 .and. saved_line ) then
        write ( u_scr2, 100 ) 'LITS ', '', ''
        write ( u_scr2, 100 ) '', errCode, trim(lineSave)
        saved_line = .false.
        cycle
      end if
      call read_line ( line, iostat )
      if ( iostat < 0 ) then  ! End of file
        if ( macro_depth == 0 ) then
          if ( saved_line ) then
            write ( u_scr2, 100 ) 'LITS ', '', ''
            write ( u_scr2, 100 ) '', errCode, trim(lineSave)
          end if
          exit o
        end if
        close ( inunit )
        macro_depth = macro_depth - 1
        inunit = inunit - 1
        if ( macro_depth < 0 ) then
          inunit = save_inunit
          if ( numCalls > 0 ) then
            call open_call
            if ( numCalls == 0 ) then
              write ( u_scr2, 100 ) 'LITS ', '', ''
              if ( saved_line ) then
                write ( u_scr2, 100 ) '', '', trim(lineSave)
                saved_line = .false.
              end if
              cycle
            end if
          end if
        end if
        cycle o
      end if
      if ( iostat > 0 ) then  ! Error
        inquire ( unit=inunit, name=file )
        call io_error ( "While reading input", inunit, file )
        stop
      end if

      why = ''
      if ( macro_depth > 0 ) why = 'GEN'

      if ( line(6:6) == '*' ) then
        write ( u_scr2, 100 ) why, errCode, trim(line)
100     format ( a5, a1, a )
      else
        select case ( line(16:20) )
        case ( 'CALL', 'INCLD' )
          operands(1)%kind = k_none
          num_Operands = 0
          i = 21
          why = 'MACRO'
          call parse ( line, i, status, .false., .false. )
          if ( operands(1)%kind /= k_symbolic ) then
            call do_error ( 'First operand of CALL or INCLD is not a name' )
            errCode = badStatement
          else
            ! Only one of each CALL or INCLD file
            do i = 1, numCalls
              if ( callQueue(i) == operands(1)%label(1:3) ) then
                numCalls = numCalls - 1
                exit
              end if
            end do
            numCalls = numCalls + 1
            callQueue(numCalls) = operands(1)%label(1:3) ! Only three chars count
          end if
          if ( numCalls >= ubound(callQueue,1) ) errCode = overcall
          call get_args
          if ( theCall > 1 .and. numArgs > 0 ) then
            call do_error ( 'Arguments not permitted within CALL file' )
            errCode = macroErr
          end if
          write ( u_scr2, 100 ) 'MACRO', errCode, trim(line)
          if ( errCode == '' .and. line(16:20) /= 'INCLD' ) then
            cmd = 'B'
            label = line(6:11)
            do i = 1, numArgs
              write ( u_scr2, 200 ) 'GEN  ', label, cmd, &
                & line(args(i-1)+2:args(i))
200           format ( a5, 6x, a6, 4x, a5, a )
              label = ''
              cmd = 'DCW'
            end do
          end if
        case ( 'END', 'EX' )
          if ( macro_depth > 0 ) then
            call do_error ( trim(line(16:20)) // ' not allowed in CALL file' )
            write ( u_scr2, 100 ) '', macroErr, trim(line)
          else
            lineSave = line ! save for processing after CALLs
            saved_line = .true.
            if ( numCalls > 0 ) call open_call
          end if
        case ( 'LTORG' )
          if ( macro_depth > 0 ) then
            call do_error ( trim(line(16:20)) // ' not allowed in CALL file' )
            write ( u_scr2, 100 ) '', macroErr, trim(line)
          else
            write ( u_scr2, 100 ) why, errCode, trim(line)
            if ( numCalls > 0 ) then
              call open_call
            else
              write ( u_scr2, 100 ) 'LITS ', '', ''
            end if
          end if
        case ( 'ORG' )
          if ( macro_depth > 0 ) then
            call do_error ( trim(line(16:20)) // ' not allowed in CALL file' )
            write ( u_scr2, 100 ) '', macroErr, trim(line)
          else
            write ( u_scr2, 100 ) why, errCode, trim(line)
          end if
        case default
!           do i = 1, size(op_codes,1)-1
!             if ( line(16:20) == op_codes(i)%op ) then
!               write ( u_scr2, 100 ) why, errCode, trim(line)
!               cycle o
!             end if
!           end do
          write ( u_scr2, 100 ) why, errCode, trim(line)
        end select
      end if
    end do o

    rewind ( u_scr2 )

  contains

    ! -------------------------------------------------  GET_ARGS  -----
    subroutine GET_ARGS
      ! Get all the args for a macro.  Pile all the necessary lines
      ! (max 5) into LINE.  Set ARGS(i) to the end of the i'th argument
      ! (ARGS(0) == 19).
      integer :: I, IOSTAT, J, NumLines
      args = 19
      i = 21
      numArgs = 0
      numLines = 1
      line(73:80) = ''
      do
        if ( line(i:i) == '' ) then ! end of args
          if ( i > 21 ) then        ! Is there a first arg?
            numArgs = numArgs + 1
            if ( numArgs > 99 ) then
              errCode = macroErr
            else
              args(numArgs) = i-1
            end if
          end if
          exit
        end if
        if ( line(i:i) == ',' ) then
          numArgs = numArgs + 1
          if ( numArgs > 99 ) then
            errCode = macroErr
            exit
          end if
          args(numArgs) = i-1
          if ( line(i+1:i+1) == ' ' ) then ! comma then blank -- need a line
            if ( numLines >= 5 ) then
              errCode = macroErr
              exit
            end if
            numLines = numLines + 1
            call read_line ( line(i+1:), iostat )
            if ( iostat < 0 ) then
              call do_error ( 'End-of-file while processing macro arguments' )
              exit
            end if
            if ( iostat > 0 ) then  ! Error
              inquire ( unit=inunit, name=file )
              call io_error ( "While reading input", inunit, file )
              stop
            end if
            write ( u_scr2, 200 ) 'MACRO', errCode, trim(line(i+1:i+80))
200         format ( a5,a1,a )
            line(i+1:i+52) = line(i+21:i+72)
            line(i+53:i+80) = ''
          end if
        else if ( line(i:i) == '@' ) then
          j = index(line,'@',back=.true.)
          if ( j == i ) then
            errCode = macroErr
            exit
          end if
          i = j
        end if
        i = i + 1
      end do
    end subroutine GET_ARGS

    ! ------------------------------------------------  OPEN_CALL  -----
    subroutine OPEN_CALL
      ! Open the next CALL file
      do
        theCall = theCall + 1
        if ( theCall > numCalls ) then
          numCalls = 0
          theCall = 0
          return
        end if
        if ( open_macro (callQueue(theCall)) < 0 ) then
          why = ''
          errCode = macroErr
          write ( u_scr2, 300 ) why, errCode, &
            &  '     *  ' // callQueue(theCall) // '  UNKNOWN'
300       format ( a5, a1, a )
        else
          inunit = u_input + macro_depth
          why = 'GEN'
          return
        end if
      end do
    end subroutine OPEN_CALL

    ! ----------------------------------------------  OPEN_MACRO  -----
    integer function Open_Macro ( Macro )
    ! Open a file with the same name as Macro, trying lower case first
    ! and then upper case.  Return the macro depth if a file is successfully
    ! opened, else -1 if no such file is found.
      use IO_UNITS, only: U_INPUT
      character(len=*), intent(in) :: Macro
      character(len=len(macro)) :: Copy ! In either lower or upper case
      integer :: I, IOSTAT
      type(path_t), pointer :: Path ! For searching
      if ( .not. associated(first_path) ) then
        allocate ( first_path )
        first_path%path = '.'
        first_path%len = 1
        last_path => first_path
      end if
      path => first_path
      macro_depth = macro_depth + 1
      open_macro = macro_depth
      do while ( associated(path) )
        do i = 1, len(macro)
          ! Try lower case first
          if ( iachar(macro(i:i)) >= iachar("A") .and. &
            &  iachar(macro(i:i)) <= iachar("Z") ) then
            copy(i:i) = achar(iachar(macro(i:i)) + iachar("a") - iachar("A"))
          else
            copy(i:i) = macro(i:i)
          end if
        end do
        open ( unit=u_input+macro_depth, status='OLD', &
          &    file=path%path(:path%len) // "/" // copy, iostat=iostat )
        if ( iostat == 0 ) return
        ! now try upper case
        do i = 1, len(macro)
          if ( iachar(macro(i:i)) >= iachar("a") .and. &
            &  iachar(macro(i:i)) <= iachar("z") ) then
            copy(i:i) = achar(iachar(macro(i:i)) + iachar("A") - iachar("a"))
          else
            copy(i:i) = macro(i:i)
          end if
        end do
        open ( unit=u_input+macro_depth, status='OLD', &
          &    file=path%path(:path%len) // "/" // copy, iostat=iostat )
        if ( iostat == 0 ) return
        path => path%next
      end do
      macro_depth = macro_depth - 1
      open_macro = -1
    end function Open_Macro

  end subroutine MACRO_PASS

end module MACRO_PASS_M
